@@ -12,7 +12,7 @@ import boto3.s3.transfer as trans
 from botocore.exceptions import ClientError, ParamValidationError
 from botocore.client import Config
 
-from .utils import read_block
+from .utils import read_block, raises, ensure_writable
 
 logger = logging.getLogger(__name__)
 
@@ -70,18 +70,22 @@ class S3FileSystem(object):
     storage.
 
     Provide credentials either explicitly (``key=``, ``secret=``) or depend
-    upon local configuration files (``~/.aws`` or ``~/.boto``).  See boto3
-    documentation for more information.
+    on boto's credential methods. See boto3 documentation for more
+    information. If no credentials are available, use ``anon=True``.
 
     Parameters
     ----------
-    anon : bool or None (default)
-        Whether to use anonymous connection (public buckets only). If None,
-        tries False and falls back to True.
+    anon : bool (False)
+        Whether to use anonymous connection (public buckets only). If False,
+        uses the key/secret given, or boto's credential resolver (environment
+        variables, config files, EC2 IAM server, in that order)
     key : string (None)
         If not anonymouns, use this key, if specified
     secret : string (None)
         If not anonymous, use this password, if specified
+    use_ssl : bool (True)
+        Whether to use SSL in connections to S3; may be faster without, but
+        insecure
     kwargs : other parameters for boto3 session
 
     Examples
@@ -95,27 +99,32 @@ class S3FileSystem(object):
     b'Hello, world!'
     """
     _conn = {}
+    _singleton = [None]
     connect_timeout = 5
     read_timeout = 15
 
-    def __init__(self, anon=None, key=None, secret=None, token=None, **kwargs):
+    def __init__(self, anon=False, key=None, secret=None, token=None,
+                 use_ssl=True, **kwargs):
         self.anon = anon
         self.key = key
         self.secret = secret
         self.token = token
         self.kwargs = kwargs
         self.dirs = {}
-        self.no_refresh = False
-        if anon is None:
-            try:
-                self.anon = False
-                self.s3 = self.connect()
-                self.ls('')
-                return
-            except:
-                logger.debug('Accredited connection failed, trying anonymous')
-                self.anon = True
+        self.use_ssl = use_ssl
         self.s3 = self.connect()
+        S3FileSystem._singleton[0] = self
+
+    @classmethod
+    def current(cls):
+        """ Return the most recently created S3FileSystem
+
+        If no S3FileSystem has been created, then create one
+        """
+        if not cls._singleton[0]:
+            return S3FileSystem()
+        else:
+            return cls._singleton[0]
 
     def connect(self, refresh=False):
         """
@@ -126,31 +135,32 @@ class S3FileSystem(object):
         refresh : bool (True)
             Whether to use cached filelists, if already read
         """
-        anon, key, secret, kwargs, token = (self.anon, self.key, self.secret,
-                                            self.kwargs, self.token)
-        tok = tokenize(anon, key, secret, kwargs, token)
+        anon, key, secret, kwargs, token, ssl = (self.anon, self.key,
+                            self.secret, self.kwargs, self.token, self.use_ssl)
+        tok = tokenize(anon, key, secret, kwargs, token, ssl)
         if refresh:
             self._conn.pop(tok, None)
+        logger.debug("Open S3 connection.  Anonymous: %s", self.anon)
+        if self.anon:
+            # TODO: test addition of kwargs (e.g., S3 data centre)
+            from botocore import UNSIGNED
+            conf = Config(connect_timeout=self.connect_timeout,
+                          read_timeout=self.read_timeout,
+                          signature_version=UNSIGNED)
+            self.session = boto3.Session(**self.kwargs)
+        else:
+            conf = Config(connect_timeout=self.connect_timeout,
+                          read_timeout=self.read_timeout)
+            self.session = boto3.Session(self.key, self.secret, self.token,
+                                         **self.kwargs)
         if tok not in self._conn:
-            logger.debug("Open S3 connection.  Anonymous: %s",
-                         self.anon)
-            if self.anon:
-                # TODO: test addition of kwargs (e.g., S3 data centre)
-                from botocore import UNSIGNED
-                conf = Config(connect_timeout=self.connect_timeout,
-                              read_timeout=self.read_timeout,
-                              signature_version=UNSIGNED)
-                s3 = boto3.Session(**self.kwargs).client('s3', config=conf)
-            else:
-                conf = Config(connect_timeout=self.connect_timeout,
-                              read_timeout=self.read_timeout)
-                s3 = boto3.Session(self.key, self.secret, self.token,
-                                   **self.kwargs).client('s3', config=conf)
+            s3 = self.session.client('s3', config=conf, use_ssl=ssl)
             self._conn[tok] = s3
         return self._conn[tok]
 
     def get_delegated_s3pars(self, exp=3600):
-        """Get temporary credentials, apropriate for sending across a network
+        """Get temporary credentials from STS, apropriate for sending across a
+        network. Only relevant where the key/secret were explicitly provided.
 
         Parameters
         ----------
@@ -166,25 +176,17 @@ class S3FileSystem(object):
         if self.token:  # already has temporary cred
             return {'key': self.key, 'secret': self.secret, 'token': self.token,
                     'anon': False}
-        sts = boto3.Session(self.key, self.secret, self.token,
-                            **self.kwargs).client('sts')
-        cred = sts.get_session_token(DurationSeconds=3600)['Credentials']
+        if self.key is None or self.secret is None:  # automatic credentials
+            return {'anon': False}
+        sts = self.session.client('sts')
+        cred = sts.get_session_token(DurationSeconds=exp)['Credentials']
         return {'key': cred['AccessKeyId'], 'secret': cred['SecretAccessKey'],
                 'token': cred['SessionToken'], 'anon': False}
 
-    def refresh_off(self):
-        """ Block auto-refresh when writing.
-
-        Use in conjunction with `refresh_on()` when writing many files to S3.
-        """
-        self.no_refresh = True
-
-    def refresh_on(self):
-        self.no_refresh = False
-
     def __getstate__(self):
         d = self.__dict__.copy()
         del d['s3']
+        del d['session']
         logger.debug("Serialize with state: %s", d)
         return d
 
@@ -209,6 +211,47 @@ class S3FileSystem(object):
                                       " and manage bytes" % (mode[0] + 'b'))
         return S3File(self, path, mode, block_size=block_size)
 
+    def _lsdir(self, path, refresh=False):
+        if path.startswith('s3://'):
+            path = path[len('s3://'):]
+        path = path.rstrip('/')
+        bucket, prefix = split_path(path)
+        prefix = prefix + '/' if prefix else ""
+        if path not in self.dirs or refresh:
+            try:
+                pag = self.s3.get_paginator('list_objects')
+                it = pag.paginate(Bucket=bucket, Prefix=prefix, Delimiter='/')
+                files = []
+                dirs = None
+                for i in it:
+                    dirs = dirs or i.get('CommonPrefixes', None)
+                    files.extend(i.get('Contents', []))
+                if dirs:
+                    files.extend([{'Key': l['Prefix'][:-1], 'Size': 0,
+                                  'StorageClass': "DIRECTORY"} for l in dirs])
+                files = [f for f in files if len(f['Key']) > len(prefix)]
+                for f in files:
+                    f['Key'] = '/'.join([bucket, f['Key']])
+            except ClientError:
+                # path not accessible
+                files = []
+            self.dirs[path] = files
+        return self.dirs[path]
+
+    def _lsbuckets(self, refresh=False):
+        if '' not in self.dirs or refresh:
+            if self.anon:
+                # cannot list buckets if not logged in
+                return []
+            files = self.s3.list_buckets()['Buckets']
+            for f in files:
+                f['Key'] = f['Name']
+                f['Size'] = 0
+                f['StorageClass'] = 'BUCKET'
+                del f['Name']
+            self.dirs[''] = files
+        return self.dirs['']
+
     def _ls(self, path, refresh=False):
         """ List files in given bucket, or list of buckets.
 
@@ -229,76 +272,65 @@ class S3FileSystem(object):
         """
         if path.startswith('s3://'):
             path = path[len('s3://'):]
-        path = path.rstrip('/')
-        bucket, key = split_path(path)
-        if bucket not in self.dirs or (refresh and not self.no_refresh):
-            if bucket == '':
-                # list of buckets
-                if self.anon:
-                    # cannot list buckets if not logged in
-                    return []
-                files = self.s3.list_buckets()['Buckets']
-                for f in files:
-                    f['Key'] = f['Name']
-                    f['Size'] = 0
-                    del f['Name']
-            else:
-                try:
-                    pag = self.s3.get_paginator('list_objects')
-                    it = pag.paginate(Bucket=bucket)
-                    files = []
-                    for i in it:
-                        files.extend(i.get('Contents', []))
-                except ClientError:
-                    # bucket not accessible
-                    raise FileNotFoundError(bucket)
-                for f in files:
-                    f['Key'] = "/".join([bucket, f['Key']])
-            self.dirs[bucket] = list(sorted(files, key=lambda x: x['Key']))
-        files = self.dirs[bucket]
-        return files
+        if path in ['', '/']:
+            return self._lsbuckets(refresh)
+        else:
+            return self._lsdir(path, refresh)
 
-    def ls(self, path, detail=False):
+    def ls(self, path, detail=False, refresh=False):
         """ List single "directory" with or without details """
         if path.startswith('s3://'):
             path = path[len('s3://'):]
         path = path.rstrip('/')
-        files = self._ls(path)
-        if path:
-            pattern = re.compile(path + '/[^/]*.$')
-            files = [f for f in files if pattern.match(f['Key']) is not None]
-            if not files:
-                try:
-                    files = [self.info(path)]
-                except (OSError, IOError, ClientError):
-                    files = []
+        files = self._ls(path, refresh=refresh)
+        if not files:
+            if split_path(path)[1]:
+                files = [self.info(path)]
+            elif path:
+                raise FileNotFoundError(path)
         if detail:
             return files
         else:
             return [f['Key'] for f in files]
 
-    def info(self, path):
+    def info(self, path, refresh=False):
         """ Detail on the specific file pointed to by path.
 
-        NB: path has trailing '/' stripped to work as `ls` does, so key
-        names that genuinely end in '/' will fail.
+        Gets details only for a specific key, directories/buckets cannot be
+        used with info.
         """
-        if path.startswith('s3://'):
-            path = path[len('s3://'):]
-        path = path.rstrip('/')
-        files = self._ls(path)
-        files = [f for f in files if f['Key'].rstrip('/') == path]
+        parent = path.rsplit('/', 1)[0]
+        files = self._lsdir(parent, refresh=refresh)
+        files = [f for f in files if f['Key'] == path and f['StorageClass'] not
+                 in ['DIRECTORY', 'BUCKET']]
         if len(files) == 1:
             return files[0]
         else:
-            raise IOError("File not found: %s" % path)
+            try:
+                bucket, key = split_path(path)
+                out = self.s3.head_object(Bucket=bucket, Key=key)
+                out = {'ETag': out['ETag'], 'Key': '/'.join([bucket, key]),
+                       'LastModified': out['LastModified'],
+                       'Size': out['ContentLength'], 'StorageClass': "STANDARD"}
+                return out
+            except (ClientError, ParamValidationError):
+                raise FileNotFoundError(path)
 
-    def walk(self, path):
-        """ Return all entries below path """
+    def _walk(self, path, refresh=False):
         if path.startswith('s3://'):
             path = path[len('s3://'):]
-        return [f['Key'] for f in self._ls(path) if f['Key'].rstrip('/'
-                ).startswith(path.rstrip('/') + '/')]
+        if path in ['', '/']:
+            raise ValueError('Cannot walk all of S3')
+        filenames = self._ls(path, refresh=refresh)[:]
+        for f in filenames[:]:
+            if f['StorageClass'] == 'DIRECTORY':
+                filenames.extend(self._walk(f['Key'], refresh))
+        return [f for f in filenames if f['StorageClass'] not in
+                ['BUCKET', 'DIRECTORY']]
+
+    def walk(self, path, refresh=False):
+        """ Return all real keys below path """
+        return [f['Key'] for f in self._walk(path, refresh)]
 
     def glob(self, path):
         """
@@ -344,14 +376,12 @@ class S3FileSystem(object):
             return {p['Key']: p['Size'] for p in files}
 
     def exists(self, path):
-        """ Does such a file exist? """
-        if path.startswith('s3://'):
-            path = path[len('s3://'):]
-        path = path.rstrip('/')
-        if split_path(path)[1]:
-            return bool(self.ls(path))
+        """ Does such a file/directory exist? """
+        bucket, key = split_path(path)
+        if key:
+            return not raises(FileNotFoundError, lambda: self.ls(path))
         else:
-            return path in self.ls('')
+            return bucket in self.ls('')
 
     def cat(self, path):
         """ Returns contents of file """
@@ -430,8 +460,7 @@ class S3FileSystem(object):
         part_info = {'Parts': parts}
         self.s3.complete_multipart_upload(Bucket=bucket, Key=key,
                     UploadId=mpu['UploadId'], MultipartUpload=part_info)
-        self._ls(bucket, refresh=True)
-
+        self.invalidate_cache(path)
 
     def copy(self, path1, path2):
         """ Copy file between locations on S3 """
@@ -442,7 +471,35 @@ class S3FileSystem(object):
                                 CopySource='/'.join([buc1, key1]))
         except (ClientError, ParamValidationError):
             raise IOError('Copy failed', (path1, path2))
-        self._ls(path2, refresh=True)
+        self.invalidate_cache(path2)
+
+    def bulk_delete(self, pathlist):
+        """
+        Remove multiple keys with one call
+
+        Parameters
+        ----------
+        pathlist : listof strings
+            The keys to remove, must all be in the same bucket.
+        """
+        if not pathlist:
+            return
+        buckets = {split_path(path)[0] for path in pathlist}
+        if len(buckets) > 1:
+            raise ValueError("Bulk delete files should refer to only one bucket")
+        bucket = buckets.pop()
+        if len(pathlist) > 1000:
+            for i in range((len(pathlist) // 1000) + 1):
+                self.bulk_delete(pathlist[i*1000:(i+1)*1000])
+            return
+        delete_keys = {'Objects': [{'Key' : split_path(path)[1]} for path
+                                   in pathlist]}
+        try:
+            self.s3.delete_objects(Bucket=bucket, Delete=delete_keys)
+            for path in pathlist:
+                self.invalidate_cache(path)
+        except ClientError:
+            raise IOError('Bulk delete failed')
 
     def rm(self, path, recursive=False):
         """
@@ -459,27 +516,35 @@ class S3FileSystem(object):
         if not self.exists(path):
             raise FileNotFoundError(path)
         if recursive:
-            for f in self.walk(path):
-                self.rm(f, recursive=False)
-            return
+            self.bulk_delete(self.walk(path))
+            if not self.exists(path):
+                return
         bucket, key = split_path(path)
         if key:
             try:
                 self.s3.delete_object(Bucket=bucket, Key=key)
             except ClientError:
                 raise IOError('Delete key failed', (bucket, key))
-            self._ls(path, refresh=True)
+            self.invalidate_cache(path)
         else:
             if not self.s3.list_objects(Bucket=bucket).get('Contents'):
                 try:
                     self.s3.delete_bucket(Bucket=bucket)
                 except ClientError:
                     raise IOError('Delete bucket failed', bucket)
-                self.dirs.pop(bucket, None)
-                self._ls('', refresh=True)
+                self.invalidate_cache(bucket)
+                self.invalidate_cache('')
             else:
                 raise IOError('Not empty', path)
 
+    def invalidate_cache(self, path=None):
+        if path is None:
+            self.dirs.clear()
+        else:
+            self.dirs.pop(path, None)
+            parent = path.rsplit('/', 1)[0]
+            self.dirs.pop(parent, None)
+
     def touch(self, path):
         """
         Create empty key
@@ -489,11 +554,12 @@ class S3FileSystem(object):
         bucket, key = split_path(path)
         if key:
             self.s3.put_object(Bucket=bucket, Key=key)
-            self._ls(bucket, refresh=True)
+            self.invalidate_cache(path)
         else:
             try:
                 self.s3.create_bucket(Bucket=bucket)
-                self._ls("", refresh=True)
+                self.invalidate_cache('')
+                self.invalidate_cache(bucket)
             except (ClientError, ParamValidationError):
                 raise IOError('Bucket create failed', path)
 
@@ -544,26 +610,6 @@ class S3FileSystem(object):
         return bytes
 
 
-@contextmanager
-def no_refresh(s3fs):
-    """ Wrap an s3fs with this to temporarily block freshing filecache on writes.
-
-    Use this if writing many small files to a bucket.
-    The filelist will only be refreshed by the next writing action, or
-    explicit call to `s3fs._ls(bucket, refresh=True)`.
-
-    Usage
-    -----
-    >>> with no_refresh(s3fs) as fs:    # doctest: +SKIP
-            [fs.touch('mybucket/file%i'%i) for i in range(1500)] # doctest: +SKIP
-    """
-    s3fs.refresh_off()
-    try:
-        yield s3fs
-    finally:
-        s3fs.refresh_on()
-
-
 class S3File(object):
     """
     Open S3 key as a file. Data is only loaded and cached on demand.
@@ -605,16 +651,13 @@ class S3File(object):
         self.end = None
         self.closed = False
         self.trim = True
+        self.mpu = None
         if mode in {'wb', 'ab'}:
             self.buffer = io.BytesIO()
             self.parts = []
             self.size = 0
             if block_size < 5 * 2 ** 20:
                 raise ValueError('Block size must be >=5MB')
-            try:
-                self.mpu = s3.s3.create_multipart_upload(Bucket=bucket, Key=key)
-            except (ClientError, ParamValidationError):
-                raise IOError('Open for write failed', path)
             self.forced = False
             if mode == 'ab' and s3.exists(path):
                 self.size = s3.info(path)['Size']
@@ -622,11 +665,16 @@ class S3File(object):
                     # existing file too small for multi-upload: download
                     self.write(s3.cat(path))
                 else:
+                    try:
+                        self.mpu = s3.s3.create_multipart_upload(Bucket=bucket, Key=key)
+                    except (ClientError, ParamValidationError) as e:
+                        raise IOError('Open for write failed', path, e)
                     self.loc = self.size
                     out = self.s3.s3.upload_part_copy(Bucket=self.bucket, Key=self.key,
                                 PartNumber=1, UploadId=self.mpu['UploadId'],
                                 CopySource=path)
-                    self.parts.append({'PartNumber': 1, 'ETag': out['CopyPartResult']['ETag']})
+                    self.parts.append({'PartNumber': 1,
+                                       'ETag': out['CopyPartResult']['ETag']})
         else:
             try:
                 self.size = self.info()['Size']
@@ -635,9 +683,7 @@ class S3File(object):
 
     def info(self):
         """ File information about this path """
-        info = self.s3.s3.head_object(Bucket=self.bucket, Key=self.key)
-        info['Size'] = info.get('ContentLength')
-        return info
+        return self.s3.info(self.path)
 
     def tell(self):
         """ Current file location """
@@ -670,11 +716,11 @@ class S3File(object):
         return self.loc
 
     def readline(self, length=-1):
-        '''
+        """
         Read and return a line from the stream.
 
         If length is specified, at most size bytes will be read.
-        '''
+        """
         self._fetch(self.loc, self.loc + 1)
         while True:
             found = self.cache[self.loc - self.start:].find(b'\n') + 1
@@ -759,7 +805,7 @@ class S3File(object):
             raise ValueError('File not in write mode')
         if self.closed:
             raise ValueError('I/O operation on closed file.')
-        out = self.buffer.write(data)
+        out = self.buffer.write(ensure_writable(data))
         self.loc += out
         if self.buffer.tell() > self.blocksize:
             self.flush()
@@ -769,15 +815,17 @@ class S3File(object):
         """
         Write buffered data to S3.
 
+        Uploads the current buffer, if it is larger than the block-size.
+
         Due to S3 multi-upload policy, you can only safely force flush to S3
         when you are finished writing.  It is unsafe to call this function
         repeatedly.
 
         Parameters
         ----------
-        force : bool (True)
-            Whether to write even if the buffer is less than the blocksize. If
-            less than the S3 part minimum (5MB), must be last block.
+        force : bool
+            When closing, write the last block even if it is smaller than
+            blocks are allowed to be.
         """
         if self.mode in {'wb', 'ab'} and not self.closed:
             if self.buffer.tell() < self.blocksize and not force:
@@ -786,29 +834,38 @@ class S3File(object):
             if self.buffer.tell() == 0:
                 # no data in the buffer to write
                 return
-            if force and self.forced and self.buffer.tell() < 5 * 2 ** 20:
-                raise IOError('Under-sized block already written')
-            if force and self.buffer.tell() < 5 * 2 ** 20:
+            if force and self.forced:
+                raise ValueError("Force flush cannot be called more than once")
+            if force:
                 self.forced = True
+
             self.buffer.seek(0)
             part = len(self.parts) + 1
             i = 0
+
+            try:
+                self.mpu = self.mpu or self.s3.s3.create_multipart_upload(
+                                Bucket=self.bucket, Key=self.key)
+            except (ClientError, ParamValidationError) as e:
+                raise IOError('Initating write failed: %s' % self.path, e)
+
             while True:
                 try:
-                    out = self.s3.s3.upload_part(Bucket=self.bucket, Key=self.key,
+                    out = self.s3.s3.upload_part(Bucket=self.bucket,
                             PartNumber=part, UploadId=self.mpu['UploadId'],
-                            Body=self.buffer.read())
+                            Body=self.buffer.read(), Key=self.key)
                     break
                 except S3_RETRYABLE_ERRORS:
                     if i < retries:
-                        logger.debug('Exception %e on S3 upload, retrying',
+                        logger.debug('Exception %e on S3 write, retrying',
                                      exc_info=True)
                         i += 1
                         continue
                     else:
-                        raise IOError('Write failed after %i retries'%retries, self)
-                except:
-                    raise IOError('Write failed', self)
+                        raise IOError('Write failed after %i retries' % retries,
+                                      self)
+                except Exception as e:
+                    raise IOError('Write failed', self, e)
             self.parts.append({'PartNumber': part, 'ETag': out['ETag']})
             self.buffer = io.BytesIO()
 
@@ -820,11 +877,10 @@ class S3File(object):
         """
         if self.closed:
             return
-        self.flush(True)
         self.cache = None
-        self.closed = True
         if self.mode in {'wb', 'ab'}:
             if self.parts:
+                self.flush(force=True)
                 part_info = {'Parts': self.parts}
                 self.s3.s3.complete_multipart_upload(Bucket=self.bucket,
                                                      Key=self.key,
@@ -832,8 +888,14 @@ class S3File(object):
                                                          'UploadId'],
                                                      MultipartUpload=part_info)
             else:
-                self.s3.s3.put_object(Bucket=self.bucket, Key=self.key)
-            self.s3._ls(self.bucket, refresh=True)
+                self.buffer.seek(0)
+                try:
+                    self.s3.s3.put_object(Bucket=self.bucket, Key=self.key,
+                                           Body=self.buffer.read())
+                except (ClientError, ParamValidationError) as e:
+                    raise IOError('Write failed: %s' % self.path, e)
+            self.s3.invalidate_cache(self.path)
+        self.closed = True
 
     def readable(self):
         """Return whether the S3File was opened for reading"""
